#include "floppy.h"
#include "../cpu/ports.h"
#include "../cpu/timer.h"
#include "../cpu/isr.h"
#include "../util/terminal.h"
#include "../libc/function.h"
#include "../libc/mem.h"

//NOTE: Some of the code is designed for handling multiple motors, but some is designed for just 1. Therefore it's only safe to use this driver to control one motor.

void fl_delay(u32 delay);
void flp_irq_wait();
void flp_irq_reset();

enum FloppyRegisters
{
   STATUS_REGISTER_A                = 0x3F0, // read-only
   STATUS_REGISTER_B                = 0x3F1, // read-only
   DIGITAL_OUTPUT_REGISTER          = 0x3F2,
   TAPE_DRIVE_REGISTER              = 0x3F3,
   MAIN_STATUS_REGISTER             = 0x3F4, // read-only
   DATARATE_SELECT_REGISTER         = 0x3F4, // write-only
   DATA_FIFO                        = 0x3F5,
   DIGITAL_INPUT_REGISTER           = 0x3F7, // read-only
   CONFIGURATION_CONTROL_REGISTER   = 0x3F7  // write-only
};

enum floppy_commands {
   CMD_SPECIFY = 3,            // SPECIFY
   CMD_WRITE_DATA = 5,         // WRITE DATA
   CMD_READ_DATA = 6,          // READ DATA
   CMD_RECALIBRATE = 7,        // RECALIBRATE
   CMD_SENSE_INTERRUPT = 8,    // SENSE INTERRUPT
   CMD_SEEK = 15,              // SEEK
};

typedef enum {
	flp_dir_read = 1,
	flp_dir_write = 2
} floppy_dir;

s8 motor_tick = -1; //Decrements once every 50ms. If it's 40, it'll take 2s to reach 0.
s8 motor = 0;

u32 fl_tick = 0;

s8 floppy_ready = 1;
s8 floppy_queue = 0; //0 -> None, 1-> Read, 2-> Write
unsigned cyl_queue = 0;
char *data_queue = 0;
void (*callback_queue)(int);

//Reserve some memory for when we read data
//Can't cross a 64k boundary ):
#define flp_dmalen 0x4800
static const char flp_dmabuf[flp_dmalen] __attribute__((aligned(0x8000)));

int flp_do_track(unsigned cyl, floppy_dir dir);

int floppy_detect_drives()
{

   port_byte_out(0x70, 0x10);
   return port_byte_in(0x71);

}

u8 SECTORS_PER_TRACK = 18;
void lba_2_chs(u32 lba, u16* cyl, u16* head, u16* sector)
{
	*cyl	= lba / (2 * SECTORS_PER_TRACK);
	*head   = ((lba % (2 * SECTORS_PER_TRACK)) / SECTORS_PER_TRACK);
	*sector = ((lba % (2 * SECTORS_PER_TRACK)) % SECTORS_PER_TRACK + 1);
}


void motor_on(int drive) //0 to 3
{
	motor_tick = 0;
	motor = drive;
	
	//Turn motor on
	port_byte_out(DIGITAL_OUTPUT_REGISTER, (0x10 << motor) + motor + 0xC);

	//Wait
	fl_delay(3);
	
}

void motor_off()
{
	motor_tick = 40; //Start 2s countdown until it actually turns off
}

void fl_delay(u32 delay)
{
	u32 start_tick = fl_tick;
	while(fl_tick < start_tick + delay);
}

int fl_write(char cmd)
{
	for(int i = 0; i < 10; i++) //500ms timeout
	{
		fl_delay(1);
		if(0x80 & port_byte_in(MAIN_STATUS_REGISTER))
		{
			//Success
			port_byte_out(DATA_FIFO, cmd);
			return 0;
		}
	}
	return 1;
}

int fl_read(unsigned char *data)
{
	for(int i = 0; i < 10; i++) //500ms to try to read
	{
		fl_delay(1); //50ms
		//Try to read
		if(0x80 & port_byte_in(MAIN_STATUS_REGISTER))
		{
			//Success!
			*data = port_byte_in(DATA_FIFO);
			return 0;
		}
	}
	//Give up
	return 1;
}

void fl_check_interrupt(int *st0, int *cyl)
{
	fl_write(CMD_SENSE_INTERRUPT);

	unsigned char t_st0; //tmps
	unsigned char t_cyl;
	
	fl_read(&t_st0);
	fl_read(&t_cyl);

	*st0 = t_st0;
	*cyl = t_cyl;
}

int fl_calibrate()
{
	int st0, cyl = -1; //Temporarily impossible values
	motor_on(0);

	for(int i = 0; i < 10; i++)
	{
		//move the head to cyl 0
		fl_write(CMD_RECALIBRATE);
		fl_write(0);

		flp_irq_wait();
		fl_check_interrupt(&st0, &cyl);

		if(st0 & 0xC0)
		{
			static char * status[] =
				{ 0, "error", "invalid", "drive" };
			print("floppy_calibrate: status = ");
			println(status[st0 >> 6]);
			continue;
        }

		if(!cyl) {
			motor_off();
			return 0;
		}
		
	}
	motor_off();
	return 1;
	
}

int fl_reset()
{
	port_byte_out(DIGITAL_OUTPUT_REGISTER, 0x00); //Disable flp controller
	port_byte_out(DIGITAL_OUTPUT_REGISTER, 0x0C); //Enable flp controller

	flp_irq_wait();

	{
		int st0, cyl;
		fl_check_interrupt(&st0, &cyl);
		UNUSED(st0);
		UNUSED(cyl);
	}
	
	//500kb/s transfer speed
	port_byte_out(CONFIGURATION_CONTROL_REGISTER, 0x00);

	fl_write(CMD_SPECIFY);
	fl_write(0xDF); //3ms steprate, 240ms unload time
	fl_write(0x02); //16ms load time, no DMA

	if(fl_calibrate()) return -1;

	return 0;
}

int flp_seek(s32 cyli, s8 head)
{
	int st0, cyl = -1; //Temporarily impossible

	motor_on(0);

	for(int i = 0; i < 10; i++)
	{
		flp_irq_reset();
		
		fl_write(CMD_SEEK);
		fl_write(head<<2);
		fl_write(cyli);

		flp_irq_wait();

		fl_check_interrupt(&st0, &cyl);

		if(st0 & 0xC0)
		{
			char * status[] = { "normal", "error", "invalid", "drive" };
			print("floppy_seek: status = ");
			println(status[st0 >> 6]);
			continue;
		}

		if(cyl == cyli)
		{
			motor_off();
			return 0;
		}
	}

	motor_off();
	println("gave up seeking");
	return -1;
}

u8 flp_int = 0;

void floppy_int_callback(registers_t regs)
{
	UNUSED(regs);
	flp_int = 1;
}

void flp_irq_reset()
{
	flp_int = 0;
}

void flp_irq_wait()
{
	while(flp_int == 0);
	flp_int = 0;
}

void floppy_loop() //Runs every 50 ms
{
	fl_tick++;
	if(motor_tick >= 0)
	{
		motor_tick--;
	}

	if(motor_tick == 0)
	{
		//Turn off motors
		port_byte_out(DIGITAL_OUTPUT_REGISTER, motor + 0xC);
		motor_tick = -1;
	}
}


void flp_main_func() //Has to be executed as a 'main' process, not by interrupts
{
	if(!floppy_ready)
	{
		if(floppy_queue == 1)
		{
			//READ
			//todo: memory copy needed?
			int code = flp_do_track(cyl_queue, flp_dir_read);
			if(code == 0)
			{
				memory_copy((unsigned char*)&flp_dmabuf, (unsigned char*)data_queue, 0x4800);
				callback_queue(0);
			}
			else
			{
				callback_queue(code);
			}
			data_queue = 0; //Clear pointer
			floppy_ready = 1;
			floppy_queue = 0;
		}
		else if(floppy_queue == 2)
		{
			//WRITE
			callback_queue(flp_do_track(cyl_queue, flp_dir_write));
			floppy_ready = 1;
			floppy_queue = 0;
		}
	}
}

int init_floppy()
{
	
	int floppies = floppy_detect_drives();
	//Only 1.44MB floppies are implemented
	//In theory, this check should never occur, because if using a different sized disk, the second stage of the bootloader will never load
	//Once other floppies are supported (if ever) there will be more code in this function. Right now it is basically just an unnecessary check.
	if (floppies == 0x40)
	{
		register_loop(floppy_loop, 50);
		register_interrupt_handler(IRQ6, floppy_int_callback);
		motor_off();
		
		return 0;
	}
	return 1;
}

static void flp_dma_init(floppy_dir dir)
{
	union
	{
		unsigned char b[4]; //4 bytes
		unsigned long l; //32 bit long
	} a, c; //Address, count
	a.l = (unsigned) &flp_dmabuf;
	c.l = (unsigned) flp_dmalen - 1; //DMA counting is weird

	//Check that address is <= 24 bits (16MB)
	//Check that count is <= 16 bits (DMA limit)
	//Check that count + addr doesn't cause a carry
	if ((a.l >> 24) || (c.l >> 16) || (((a.l&0xffff) + c.l) >> 16))
	{
		println("CRITICAL: Floppy DMA buffer problem!");
	}

	unsigned char mode;
	switch (dir)
	{
		// 01 0 0 01 10 = single/inc/no-auto/to-mem/chan2
	case flp_dir_read:
		mode = 0x46;
		break;

		// 01 0 0 10 10 = single/inc/no-auto/from-mem/chan2
	case flp_dir_write:
		mode = 0x4a;
		break;
		
	default:
		println("CRITICAL: Invalid floppy direction specified.");
	}

	port_byte_out(0x0a, 0x06); //Mask chan 2

	port_byte_out(0x0c, 0xff); //Reset flipflop
	port_byte_out(0x04, a.b[0]); //Address low byte
	port_byte_out(0x04, a.b[1]); //Address high byte

	port_byte_out(0x81, a.b[2]);

	port_byte_out(0x0c, 0xff); //Reset flipflop
	port_byte_out(0x05, c.b[0]); //Count low byte
	port_byte_out(0x05, c.b[1]); //Count high byte

	port_byte_out(0x0b, mode); //Set mode based on above switch
	
	port_byte_out(0x0a, 0x02); //Unmask chan 2
}

//The function that actually does everything
int flp_do_track(unsigned cyl, floppy_dir dir)
{
	// transfer command, set below
    unsigned char cmd;

    // Read is MT:MF:SK:0:0:1:1:0, write MT:MF:0:0:1:0:1
    // where MT = multitrack, MF = MFM mode, SK = skip deleted
    // 
    // Specify multitrack and MFM mode
    static const int flags = 0xC0;
    switch(dir) {
        case flp_dir_read:
            cmd = CMD_READ_DATA | flags;
            break;
        case flp_dir_write:
            cmd = CMD_WRITE_DATA | flags;
            break;
        default: 

            println("CRITICAL: floppy_do_track: invalid direction");
            return 0; // not reached, but pleases "cmd used uninitialized"
    }

    // seek both heads
    if(flp_seek(cyl, 0)) return -1;
    if(flp_seek(cyl, 1)) return -1;

    int i;
    for(i = 0; i < 20; i++) {
        motor_on(0);

        // init dma..
        flp_dma_init(dir);

        fl_delay(2); // give some time (100ms) to settle after the seeks

		flp_irq_reset();

        fl_write(cmd);  // set above for current direction
        fl_write(0);    // 0:0:0:0:0:HD:US1:US0 = head and drive
        fl_write(cyl);  // cylinder
        fl_write(0);    // first head (should match with above)
        fl_write(1);    // first sector, strangely counts from 1
        fl_write(2);    // bytes/sector, 128*2^x (x=2 -> 512)
        fl_write(18);   // number of tracks to operate on
        fl_write(0x1b); // GAP3 length, 27 is default for 3.5"
        fl_write(0xff); // data length (0xff if B/S != 0)
        
        flp_irq_wait(); // don't SENSE_INTERRUPT here!

        // first read status information
		//IMPORTANT: All of these fl_reads might be broken(casting?)
        unsigned char st0, st1, st2, rcy, rhe, rse, bps;
        fl_read(&st0);
        fl_read(&st1);
        fl_read(&st2);
        /*
         * These are cylinder/head/sector values, updated with some
         * rather bizarre logic, that I would like to understand.
         *
         */
        fl_read(&rcy);
		fl_read(&rhe);
		fl_read(&rse);
        // bytes per sector, should be what we programmed in
        fl_read(&bps);

        int error = 0;

        if(st0 & 0xC0) {
            char * status[] = { 0, "error", "invalid command", "drive not ready" };
            print("floppy_do_sector: status = ");
			println(status[st0 >> 6]);
            error = 1;
        }
        if(st1 & 0x80) {
            print("floppy_do_sector: end of cylinder\n");
            error = 1;
        }
        if(st0 & 0x08) {
            print("floppy_do_sector: drive not ready\n");
            error = 1;
        }
        if(st1 & 0x20) {
            print("floppy_do_sector: CRC error\n");
            error = 1;
        }
        if(st1 & 0x10) {
            print("floppy_do_sector: controller timeout\n");
            error = 1;
        }
        if(st1 & 0x04) {
            print("floppy_do_sector: no data found\n");
            error = 1;
        }
        if((st1|st2) & 0x01) {
            print("floppy_do_sector: no address mark found\n");
            error = 1;
        }
        if(st2 & 0x40) {
            print("floppy_do_sector: deleted address mark\n");
            error = 1;
        }
        if(st2 & 0x20) {
            print("floppy_do_sector: CRC error in data\n");
            error = 1;
        }
        if(st2 & 0x10) {
            print("floppy_do_sector: wrong cylinder\n");
            error = 1;
        }
        if(st2 & 0x04) {
            print("floppy_do_sector: uPD765 sector not found\n");
            error = 1;
        }
        if(st2 & 0x02) {
            print("floppy_do_sector: bad cylinder\n");
            error = 1;
        }
        if(bps != 0x2) {
            print("floppy_do_sector: wanted 512B/sector");
            error = 1;
        }
        if(st1 & 0x02) {
            print("floppy_do_sector: not writable\n");
            error = 2;
        }

        if(!error) {
            motor_off();
            return 0;
        }
        if(error > 1) {
            print("floppy_do_sector: not retrying..\n");
            motor_off();
            return -2;
        }
    }

    print("floppy_do_sector: 20 retries exhausted\n");
    motor_off();
    return -1;

}

int floppy_read_track(unsigned cyl, char *data, void(*callback)(int response)) //TODO: should call callback func
{
	if(floppy_ready)
	{
		cyl_queue = cyl;
		callback_queue = callback;
		data_queue = data;
		floppy_ready = 0;
		floppy_queue = 1;
	}
	return 27; //Data transfer already in progress
}

int floppy_write_track(unsigned cyl, char *data, void (*callback)(int response))
{
	if(floppy_ready)
	{
		memory_copy((unsigned char*)data, (unsigned char*)&flp_dmabuf, 0x4800);
		cyl_queue = cyl;
		callback_queue = callback;
		floppy_ready = 0;
		floppy_queue = 2;
		return 0;
	}
	return 27; //Data transfer already in progress
}
